#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals 		= 3
#pragma IgorVersion		= 6.36
#pragma ModuleName		= SpectraImageProfile
#pragma version			= 3.90
#pragma DefaultTab		= {3,20,4}		// set default tab width in Igor Pro 9 and later

static Constant kVersion = 3.90
static StrConstant kVersionDate = "03/2022"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21549																// the project node on IgorExchange
static StrConstant ksShortTitle = "Profile Generator"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//
//	SpectraProfile is a environment to create 1D profiles from 2D data sets in horizontal or vertical
//	direction. Various modifications to the 2D data itself can be applied. Some functionality relies on
//	properties of the data used (such as scaling and column labels) and can be changed below.
//	
// 2021-02-16 - ver. 3.60:	Initial public release.
// 2021-07-22 - ver. 3.70:	Slightly adjusted the names in the main menu entries.
// 							Fixed wrong wave labels (row vs. col) for the split tool.
//							Added scaled positions to the wave name for the split tool.
//							Added keyboard shortcuts for Normalize Profile and Snap to Grid.
//							Added Accept & Quit button.
//							Added optional logging of data edits in the history.
//							Added control to set the smoothing factor for the align feature.
//							Moved all controls into an external panel.
//							Now the live profile under the mouse cursor gets normalized as well.
//							Fixed the way-too-big smoothing value when using the align feature.
//							Data units get displayed in the graph now.
// 2021-07-24 - ver. 3.71:	Improvements of matrix calculations thanks to suggestions by A.G. (WaveMetrics).
//							Added image contrast controls.
// 2021-10-26 - ver. 3.72:	When combining waves the source waves are recorded in the combined wave's notes now.
//							Added sub-step shifting to the Align feature.
// 2021-12-23 - ver. 3.80:	Made sure that sub-step alignment does not throw errors if the fit fails.
//							Fixed wrong panel size on windows with big fonts.
//							Align features with cursors present was not working correctly in some cases.
//							Added Align Edges feature which aligns features in the derivative.
//							Image profile now works with 3D data as well, either on the current layer or the merged image.
// 2022-03-23 - ver. 3.90:	Fixed error message when QuickProfile and QuickSplit are run with a folder selected.
//							Reduced ROI expansion via keyboard to single steps (unless snap-to-grid is activated).
//							Steps for repositioning and width changes via keyboard are now adapted to the data size.
//							Initial center value will be aligned to the grid for waves with less than 100 rows / columns, and honors zero.
//							Added Delete button to delete individual layers from 3D data.
//							Integration ranges are now indicated in the output name with 'to' instead of a hyphen/minus to avoid confusion.
//							Now multiple instances of the tool can be opened at the same time.
//							Added support for keyboard-shortcut (and other future) settings via external text file.
//							Axis split between image and profile can now be dragged to a new position.
//________________________________________________________________________________________________

static StrConstant kWorkingDir	 = "Profile_Generator"											// the name for the working folder
static StrConstant kSettingsFile = "Spectra Tools settings.dat"									// just for keyboard shortcuts for now.
static Constant	kNormSmooth	 = 2		// profile normalization: integer factor for smoothing the profile before division (0 = no smoothing)
static Constant	kAlignSmooth = 2		// only for backwards compatibility - align features: smoothing factor in percent of the wave's x/y dimension (e.g., 2 => points * 0.02)

//-------------------------------- functions for user code ---------------------------------------
static StrConstant kUserInfoBtnName = "Ring Current"											// the title of the user button

static Function Profile_ActivateUserInfoButton(inwave)											// called to check whether or not the data fulfills conditions for measurement normalization
	Wave inwave
	Variable activate = 0
	// +++++++++++++++++++++++++++ conditions for activation +++++++++++++++++++++++++++++++++++++
	activate = StringMatch(GetDimLabel(inwave,1,0),"*mA*")										// check if dim-labels contain "mA" values
	activate = StringMatch(note(inwave),"*ring current normalized*")? 0 : activate				// already normalized?
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return activate																				// return 1 to activate button
End

static Function Profile_ExecuteUserInfoButton(inwave)											// normalization of the data to measurement info (when the button is pressed)
	Wave inwave
	Variable valAvg = 0
	// ++++++++++++++++++++++ ring current normalization procedure +++++++++++++++++++++++++++++++
	Variable i, value, dim = WaveDims(inwave)
	for (i = 0; i < DimSize(inwave,dim-1); i += 1)
		String Currlabel = GetDimLabel(inwave,1,i)												// get the information from the column label
		sscanf Currlabel, "%s (%f mA)", Currlabel, value										// search for ring current info in the form "time (XXX mA)"
		if (value != 0)
			if (dim == 3)
				inwave[][][i] /= value
			else
				inwave[][i] /= value
			endif
		endif
		valAvg += value																			// add the values for intensity re-normalization later
	endfor
	valAvg /= i+1
	Note inwave, "ring current normalized"														// write into note that the normalization has been done
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return valAvg																				// return the intensity normalization factor which may be used restore the average height
End
//------------------------------------------------------------------------------------------------

Menu "Spectra Tools"
	SpectraImageProfile#MenuEntry("start_GUI")	,/Q, CreateBrowser; SpectraImageProfile#LoadWrapper()
	"-"
	SpectraImageProfile#MenuEntry("append_2D")	,/Q, CreateBrowser; CombineDataSets()
	SpectraImageProfile#MenuEntry("stack_2D")	,/Q, CreateBrowser; CombineDataSets(stack3D=1)
	"-"
	SpectraImageProfile#MenuEntry("profile_H")	,/Q, CreateBrowser; QuickProfileWrapper(0)
	SpectraImageProfile#MenuEntry("profile_V")	,/Q, CreateBrowser; QuickProfileWrapper(1)
	SpectraImageProfile#MenuEntry("profile_L")	,/Q, CreateBrowser; QuickProfileWrapper(2)
	"-"
	SpectraImageProfile#MenuEntry("split_H")	,/Q, CreateBrowser; QImageSplitWrapper(0)
	SpectraImageProfile#MenuEntry("split_V")	,/Q, CreateBrowser; QImageSplitWrapper(1)
	SpectraImageProfile#MenuEntry("split_L")	,/Q, CreateBrowser; QImageSplitWrapper(2)
End

//________________________________________________________________________________________________
//	Quick Profile - averages 2D/3D waves in one direction (along x, y or z) for data reduction
//________________________________________________________________________________________________

Function QuickProfileWrapper(dim)
	Variable dim
	Variable index=0
	do
		String name = GetBrowserSelection(index)
		if (strlen(name) <= 0)
			break
		endif
		Wave/Z work = $name
		index += 1
		if (!WaveExists(work))
			continue
		endif
		QProfile(work,dim)	
	while(1)
	return 0
End
//################################################################################################
Function QProfile(inwave,dim)
	Wave inwave
	Variable dim 		// 0 = H, 1 = V, 2 = L
	
	Variable wdim = WaveDims(inwave)
	String name = NameOfWave(inwave)
	
	dim = round(dim)
	if (dim < 0 || dim > 2)
		Abort "Wrong dimension parameter. Must be 0, 1 or 2"
	endif
	if (wdim < 2 && dim == 1)
		Abort "This is meant for 2D waves!"
	endif
	if (wdim != 3 && dim == 2)
		Abort "This is meant for 3D waves!"
	endif

	Variable chars = IgorVersion() < 7.00 ? 26 : inf
	String outName = name[0,chars] + SelectString(dim-1, "_prH", "_prV","_prL")					// the name of the output wave
	if (dim==0)
		MatrixOP/O $outName = sumRows(inwave)/numCols(inwave)
	elseif (dim==1)
		MatrixOP/O $outName = sumRows(inwave^T)/numRows(inwave)
	elseif (dim==2)
		MatrixOP/O $outName = sumBeams(inwave)/waveLayers(inwave)
	endif
	Wave outwave = $outName
	
	if (wdim == 3 && dim < 2)																	// for profiles of 3D waves -> rotate result to 2D map
		MatrixOP/FREE temp = transposeVol(outwave,1)
		Duplicate/O temp, outwave
		SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), outwave
	endif
	
	String notes = note(inwave) + "\r" + name													// scale & write a note of what has been done
	if (dim < 2)
		SetScale/P x, DimOffset(inwave,dim), DimDelta(inwave,dim), WaveUnits(inwave,dim), outwave
		notes += " integrated from "+num2str(DimOffset(inwave,1-dim))+" to "+num2str(DimOffset(inwave,1-dim)+(DimSize(inwave,1-dim)-1)*DimDelta(inwave,1-dim))+" "+WaveUnits(inwave,1-dim)
	elseif (dim == 2)
		CopyScales inwave, outwave
		notes += " integrated over all layers."
	endif
	Note outwave, notes
	
	return 0
End

//________________________________________________________________________________________________
//	Image split - splits 2D/3D data into individual waves along x, y or z
//________________________________________________________________________________________________

Function QImageSplitWrapper(dim)
	Variable dim
	Variable index=0
	do
		String name = GetBrowserSelection(index)
		if (strlen(name) <= 0)
			break
		endif
		Wave work = $name
		index += 1
		if (!WaveExists(work))
			continue
		endif
		QImageSplit(work,dim)	
	while(1)
	return 0
End
//################################################################################################
Function QImageSplit(inwave,dim)
	Wave inwave
	Variable dim 					// 0 = H, 1 = V, 2 = L
	
	Variable wdim = WaveDims(inwave)
	String name = NameOfWave(inwave)
	
	dim = round(dim)
	if (dim < 0 || dim > 2)
		Abort "Wrong dimension parameter. Must be 0, 1 or 2"
	endif
	if (wdim < 2 && dim == 1)
		Abort "This is meant for 2D or 3D waves!"
	endif
	if (wdim != 3 && dim == 2)
		Abort "This is meant for 3D waves!"
	endif
	
	DFREF saveDFR = GetDataFolderDFR()
	NewDataFolder/O/S $name
	
	Variable i
	String wname, unit, scaledpos
	Switch (dim)
		case 0:
			unit = WaveUnits(inwave,1)
			for (i = 0; i < DimSize(inwave, 1); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,1)+i*DimDelta(inwave,1)) + SelectString(strlen(unit),""," ") + unit + ")"
				if (wdim == 3)
					wname = "col_img_" + num2str(i) + scaledpos
					MatrixOP/Free temp = col(inWave,i)
					MatrixOP/O $wname = transposeVol(temp,1)
					SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
					SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), $wname
				else
					wname = "col_" + num2str(i) + scaledpos
					MatrixOP/O $wname = col(inWave,i)
					SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
				endif
			endfor
		break
		case 1:
			unit = WaveUnits(inwave,0)
			for (i = 0; i < DimSize(inwave, 0); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,0)+i*DimDelta(inwave,0)) + SelectString(strlen(unit),""," ") + unit + ")"
				if (wdim == 3)
					wname = "row_img_" + num2str(i) + scaledpos
					MatrixOP/Free temp = row(inWave,i)
					MatrixOP/O $wname = transposeVol(temp,4)
					SetScale/P x, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
					SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), $wname
				else
					wname = "row_" + num2str(i) + scaledpos
					MatrixOP/O $wname = row(inWave,i)^t
					SetScale/P x, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
				endif
			endfor
		break
		case 2:
			unit = WaveUnits(inwave,2)
			for (i = 0; i < DimSize(inwave, 2); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,2)+i*DimDelta(inwave,2)) + SelectString(strlen(unit),""," ") + unit + ")"
				wname = "layer_" + num2str(i) + scaledpos
				MatrixOP/O $wname=inWave[][][i]
				SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
				SetScale/P y, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
			endfor
		break
	EndSwitch
	SetDataFolder saveDFR
	
	return 0
End

//________________________________________________________________________________________________
//	Combines individual 1D or 2D datasets to one 2D/3D dataset
//________________________________________________________________________________________________

Function CombineDataSets([stack3D])
	Variable stack3D																			// stack3D = 1 concatenates the data without summing
	
	String name = "", list = "", waves= ""
	Variable index = 0, items
	do																							// make a list of the selected data
		name = GetBrowserSelection(index)
		if (strlen(name) <= 0)
			break
		endif
		list += name + ";"
		waves+= ParseFilePath(0, name, ":", 1, 0) + ";"
		index+=1
	while(1)
	list  = Sortlist(list, ";", 16)
	items = itemsinlist(list)
	
	// +++++++++++++++++++++++++++++++++ build an output name ++++++++++++++++++++++++++++++++++++
	String number1, number2, outname
	name = ParseFilePath(0, StringFromList(items-1, list), ":", 1, 0)							// try to get the last data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number2													// get the end number in the name
	name = ParseFilePath(0, StringFromList(0, list), ":", 1, 0)									// try to get the first data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number1
	
	outname = RemoveEnding(ReplaceString("'",name,""), number1)
	outname = RemoveEnding(outname, "_")
	if(strlen(number1) > 0 && strlen (number2) > 0)												// add range to name if both numbers are set
		outname += "_"+number1+"-"+number2
	endif
	Variable chars = IgorVersion() < 7.00 ? 26 : inf
	outname = outname[0,chars] + "_cmb"
	
	if (!ParamIsDefault(stack3D) && stack3D == 1)
		Concatenate/O list, $outname
		return 0																				// stop here
	endif
	
	// ++++++++++++++++++++++++++++ continue when stack3D is not set +++++++++++++++++++++++++++++
	Make/FREE/N=(items) xSizes, ySizes, Offset, Deltas, RngEnd									// temporary data holding all the scaling and unit values from all waves
	Make/FREE/N=(items)/T xUnits, yUnits
	
	Wave work 		= $StringFromList(0, list)													// extract y-scaling and names from first wave
	Variable yOffset= DimOffset(work, 1)
	Variable yDelta	= DimDelta(work, 1)
	String yUnit	= WaveUnits(work, 1)
	String wname	= NameOfWave(work)
	String notes	= note(work)
	
	For (index = 0; index < items; index += 1)
		Wave work = $StringFromList(index, list)
		xSizes[index] = DimSize(work, 0)
		ySizes[index] = DimSize(work, 1) > 0 ? DimSize(work, 1) : 1
		Offset[index] = DimOffset(work, 0)
		Deltas[index] = DimDelta(work, 0)
		RngEnd[index] = Offset[index] + Deltas[index] * (xSizes[index]-1)
		xUnits[index] = WaveUnits(work, 0)
	Endfor
	
	Variable Delta = Deltas[0], DeltaMisMatch = 0
	if (Delta != mean(Deltas))																	// may lead to data overrun
		DoAlert 1, "Data x spacing does not match! Do you want to use interpolation? Press no to exit."
		if (V_flag == 1)
			Delta = Delta > 0 ? WaveMin(Deltas) : WaveMax(Deltas)
			DeltaMisMatch = 1
		else
			return 0
		endif
	endif
	
	Variable WaveStart	= Delta > 0? WaveMin(Offset) : WaveMax(Offset)
	Variable WaveEnd	= Delta > 0? WaveMax(RngEnd) : WaveMin(RngEnd)
	Variable xSize		= WaveMax(xSizes)
	if (DeltaMisMatch)
		xSize = abs((WaveStart-WaveEnd)/Delta)+1
	endif
	
	Make/O/D/N=(xSize, Sum(Ysizes)) $outname = 0
	Wave out = $outname
	SetScale/I x, WaveStart, WaveEnd, xUnits[0], out
	SetScale/P y, yOffset, yDelta, yUnit, out
	Note out, notes + "\rCombined data of "+ReplaceString(";",RemoveEnding(waves),", ")

	Variable startpos = 0, endpos, startpnt, endpnt
	For (index = 0; index < itemsinlist(list); index += 1)
		endpos	  = startpos + ySizes[index]-1
		WaveStart = Offset[index]
		Wave work = $StringFromList(index, list)

		startpnt  = x2pnt(out, pnt2x(work, 0))													// find the sub-range of valid x-values from work
		endpnt	  = x2pnt(out, pnt2x(work, xSizes[index]-1))
		startpnt += (pnt2x(work, 0) - pnt2x(out, startpnt) > 1e-6)								// make sure the x-range is not even slightly outside the valid range
		endpnt	 -= (pnt2x(out, endpnt) - pnt2x(work, xSizes[index]-1) > 1e-6)	

		out[startpnt,endpnt][startpos,endpos] = work(x)[q-startpos] 
		startpos	+= ySizes[index]
	Endfor
	
	return 0
End

//________________________________________________________________________________________________
//	Generate profiles from 2D or 3D waves
//________________________________________________________________________________________________

static Function LoadWrapper()
	Wave/Z work = $GetBrowserSelection(0)	// first selection
	if (WaveExists(work))
		SpectraProfile(work)
	endif
	return 0
End

//################################################################################################

static Function/S MenuEntry(which)
	String which
	Variable fileID

	String read = "", menuStr = "", searchStr = ""
	StrSwitch (which)
		case "start_GUI":
			menuStr = "Image Profile ..."
			searchStr = "start image profile"
		break
		case "append_2D":
			menuStr = "Append 1D,2D Datasets into 2D Data"
			searchStr = "append 2D data"
		break
		case "stack_2D":
			menuStr = "Stack 2D Datasets into 3D Data"
			searchStr = "stack 2D data"
		break
		case "profile_H":
			menuStr = "Profile of 2D|3D Data - [Horizontal]"
			searchStr = "quick horizontal profile"
		break
		case "profile_V":
			menuStr = "Profile of 2D|3D Data - [Vertical]"
			searchStr = "quick vertical profile"
		break
		case "profile_L":
			menuStr = "2D Image from 3D Data - [Layers]"
			searchStr = "quick layer profile"
		break
		case "split_H":
			menuStr = "Split Up 2D|3D Data - [Horizontal]"
			searchStr = "data split horizontal"
		break
		case "split_V":
			menuStr = "Split Up 2D|3D Data - [Vertical]"
			searchStr = "data split vertical"
		break
		case "split_L":
			menuStr = "Split Up 3D Data - [Layers]"
			searchStr = "data split layers"
		break
		default:
			return ""
	EndSwitch
	
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFile
	if (!V_flag)
		do
			FReadLine fileID, read
		while (strlen(read) > 0 && !StringMatch(read, searchStr+"*"))
		read = ReplaceString("\t", ReplaceString(" ", read, ""), "")
		read = StringByKey(ReplaceString(" ", searchStr, ""),read,"=","\r")
		Close fileID
	endif
	return menuStr+SelectString(strlen(read),"","/"+read)
End

//################################################################################################

Function SpectraProfile(inwave)
	Wave/Z inwave
	if (!WaveExists(inwave))
		return -1
	endif
	Variable dim = WaveDims(inwave)
	if (dim < 2 || dim > 3 || WaveType(inwave,1) != 1 )
		Abort "This tool works only for numeric 2D or 3D waves."
	endif
	//+++++++++++++++++++++++++++++++++ generate working variables ++++++++++++++++++++++++++++++
	String gTitle = "Spectra Image Profile (ver. "+num2str(kVersion)+" - "+kVersionDate+"): Profile of "
	String gName = UniqueName("ProfileGeneratorGraph", 6, 1), pName = "ImageProfilePanel"
	String pPath = gName+"#"+pName
	String workF = "IP"+ReplaceString("ProfileGeneratorGraph",gName,"")+"_"+CleanupName(NameOfWave(inwave),0)[0,25]
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:$(kWorkingDir)
	NewDataFolder/O root:Packages:$(kWorkingDir):$(workF)										// the folder for temporary stuff
	DFREF stg = root:Packages:$(kWorkingDir):$(workF)
	
	String/G		stg:InwaveLocation = GetWavesDataFolder(inwave,2)							// save wave name and folder for later
	Variable/G		stg:ProfileUpdateConnector													// formula connection variable... has no purpose itself, but is an anchor for triggering updates
	Make/O/D 		stg:SectionLine																// the currently displayed profile
	Make/O/D/N=5	stg:CurrentLines = 0														// holds the range values ([0] = current working mode {0 = horizontal, 1 = vertical}, [1] = h.pos, [2] = h.width, [3] = v.pos, [4] = v.width)
	Make/O/D/N=5 	stg:LinesY, stg:LinesX														// marquee line waves for the image 
	SetScale d 0,0,WaveUnits(inwave,0), stg:LinesX												// the data units will be displayed on the x- and y-axes
	SetScale d 0,0,WaveUnits(inwave,1), stg:LinesY
	Duplicate/O		inwave, stg:StartCopy														// copy original wave for undo

	NewDataFolder/O	root:WinGlobals
	NewDataFolder/O	root:WinGlobals:$(gName)													// here all main window related data is generated
	String/G		root:WinGlobals:$(gName):S_TraceOffsetInfo = ""								// Igor defined window status parameter
	//+++++++++++++++++++++++++++++++++ generate the graph ++++++++++++++++++++++++++++++++++++++
	Wave LinesX = stg:LinesX
	Wave LinesY	= stg:LinesY
	Variable pix = 72/ScreenResolution
	Variable TFTResX= Str2Num(StringFromList(3,StringByKey("SCREEN1",IgorInfo(0),":"),","))		// get the display resolution to align window better (center of full screen)
	Variable TFTResY= Str2Num(StringFromList(4,StringByKey("SCREEN1",IgorInfo(0),":"),","))
	Variable WinTop = 70, WinLeft = 200
	Variable WinBottom	= TFTResY - 150
	Variable WinRight	= TFTResX - 350

	Display/K=2/W=(WinLeft*pix, WinTop*pix, WinRight*pix, WinBottom*pix)/N=$gName as gTitle+NameOfWave(inwave)
	AppendImage/G=1 inwave																		// don't append as RGB image
	ModifyImage ''#0 ctab= {*,*,Terrain,0}
	AppendToGraph/C=(0,0,0) LinesY vs LinesX													// add marquee lines
	ModifyGraph quickdrag(LinesY)=1,live(LinesY)=1												// lines can be dragged (act on S_TraceOffsetInfo)
	
	SetWindow $gName hook(CreateProfile)=Profile_ProcessWindowEvent								// keeps track of window changes
	NVar ProfileUpdateConnector = stg:ProfileUpdateConnector									// set formula for live update
	SetFormula ProfileUpdateConnector,"LinesIntegrateUpdate(root:WinGlobals:"+gName+":S_TraceOffsetInfo)"
	//++++++++++++++++++++++++++++++++++++ now the controls ++++++++++++++++++++++++++++++++++++++
	Variable panelW = 270
	Variable panelH = 590 + 60 * (dim == 3)
	
	NewPanel/HOST=$gName/EXT=0/K=2/N=$pName/W=(0, 0, panelW, panelH) as "Spectra Profile"
	SetWindow $pPath hook(CreateProfile)=Profile_ProcessWindowEvent								// keeps track of window changes here as well (only button presses)
	SetWindow $pPath userdata(workFolder)=workF													// save work folder location
	SetWindow $gName userdata(procVersion)=num2str(kVersion)									// save procedure version
	
	TitleBox IntMode			,pos={10,10}	,size={80,23}	,title="Mode:"	 			,frame=0	,fstyle=1	,fSize=12
	CheckBox IntColCheckBox		,pos={60,10}	,size={80,23}	,title="\f04H\f00orizontal: ="		,help={"Select horizontal mode. (key: 'h')"}
	CheckBox IntRowCheckBox		,pos={165,10}	,size={80,23}	,title="\f04V\f00ertical: ||"		,help={"Select vertical mode. (key: 'v')"}
	SetVariable Position		,pos={10,33}	,size={150,23}	,title="Center:"					,help={"Center position of the selection. Move with cursor keys."}	
	SetVariable Width			,pos={10,58}	,size={150,23}	,title="Width:"						,help={"Width of the selection. This is the distance between the two lines. Move with cursor keys."}
	CheckBox FullWidth			,pos={170,34}	,size={70,23}	,title="Full Size"					,help={"Select the full range of the spectrum. (key: page up)"}
	CheckBox MinWidth			,pos={170,59}	,size={70,23}	,title="Minimal Size"				,help={"Select one slice. (key: page down)"}
	
	Button bSaveSpec			,pos={20,90}	,size={100,26}	,title="\f04S\f00ave Profile"		,help={"Save current profile. (key: 's') Press 'shift' to omit range in the wave name."}
	Button bCompare				,pos={20,125}	,size={100,26}	,title="Sna\f04p\f00shot"			,help={"Click to append and compare current profile. (key: 'p')"}
	
	CheckBox Normto1			,pos={140,90}	,size={80,23}	,title="\f04N\f00ormalize Profile"	,help={"Normalizes the profile to 1. (key: 'n')"}
	CheckBox MousetoSpec		,pos={140,112}	,size={80,23}	,title="\f04M\f00ouse Csr Profile"	,help={"Live profile under the mouse cursor. (key: 'm')"}
	CheckBox SnapToGrid			,pos={140,134}	,size={80,23}	,title="Snap Lines to \f04G\f00rid"	,help={"Snap the selection lines to full rows/columns. (key: 'g')"}

	CheckBox cPrintEdits		,pos={22,161}	,size={250,23}	,title="Print a Log of Data Edits to History"				,fstyle=2

	GroupBox NormalizeGroup		,pos={10,185}	,size={250,83}	,title="Normalize 2D Data / Current Layer"		,frame=0	,fstyle=1
	CheckBox MaintainInt		,pos={22,210}	,size={80,23}	,title="Keep average Int."			,help={"For normalization: Check to maintain the (average) height of the spectrum. Otherwise the spectrum will be scaled down while normalizing."}	
	Button bNormalize			,pos={20,235}	,size={110,23}	,title="Div. by Profile"			,help={"Normalize the image or selected layer by the current profile (divide). Holding 'ctrl / command' disables smoothing (hard normalize)."}
	Button bNormUser			,pos={140,205}	,size={110,23}	,title=kUserInfoBtnName				,help={"Normalize the image or selected layer with the included measurement information, e.g., ring current."}
	Button bNormMax				,pos={140,235}	,size={110,23}	,title="Each Maximum"				,help={"Normalize the image or selected layer to the maximum of every horizontal / vertical line."}
	
	GroupBox ModifyGroup		,pos={10,275}	,size={250,235}	,title="Edit Data"			,frame=0	,fstyle=1
	Button bSubtractPr			,pos={20,300}	,size={130,23}	,title="Subtract Profile"			,help={"Cut the current Profile from the image or current layer. Press and hold 'ctrl / command' (or 'alt / option') to subtract along features moving with increasing (decreasing) Y values."}
	Button bRemoveSelection		,pos={20,330}	,size={130,23}	,title="Delete ROI Section"			,help={"Remove the selected data between the lines from the image or ALL layers of 3D data. (key: del)"}
	Button bShiftYAxisPlus		,pos={160,300}	,size={90,23}	,title="> Shift: +Y"				,help={"Shifts 2D / 3D data by adding Y-axis values to X values. Hold the 'ctrl / command' key, to cut image to exclude partial data."}
	Button bShiftYAxisMinus		,pos={160,330}	,size={90,23}	,title="< Shift: -Y"				,help={"Shifts 2D / 3D data by subtracting Y-axis values from X values. Hold the 'ctrl / command' key, to cut image to exclude partial data."}

	GroupBox SeparationLine1	,pos={20,359}	,size={228,1}	,frame=0
	Button bAlign				,pos={20,371}	,size={65,55}	,title="Align\nFeatures"			,help={"Use data correlation to align features in the image or current layer (press 'shift' for top to bottom alignment). Set cursors to align to a sub-region!"}
	CheckBox cAlignSmooth		,pos={95,370}	,size={80,23}	,title="Smoothing:"					,help={"Smooth profile before trying alignment (helps with low S/N)."}
	CheckBox cAlignOutput		,pos={95,412}	,size={80,23}	,title="Result Output"				,help={"Saves the used shift value for each line alignment in a wave. This can be used to log the amount of shift or analyze this result further."}
	CheckBox cAlignSubShift		,pos={185,391}	,size={80,23}	,title="Substeps"					,help={"Shifts data in sub-steps if possible (interpolation is used)."}
	CheckBox cAlignEdges		,pos={95,391}	,size={80,23}	,title="Align Edges"				,help={"Align edges instead of peak features (using the derivative)."}
	SetVariable SmoothFactor	,pos={140,368}	,size={110,23}	,title=""							,help={"The smoothing factor (no. of binomial smoothing operations) applied before alignment."}

	GroupBox SeparationLine2	,pos={20,435}	,size={228,1}	,frame=0
	Button bUndoAll				,pos={20,445}	,size={150,23}	,title="\f05U\f01ndo All Changes"	,help={"Restore the original wave. (key: 'u')"} ,fSize=13
	Button bUndoLast			,pos={20,475}	,size={150,23}	,title="Undo Last Change"			,help={"Restore the data before the last action."}
	Button bSaveQuit			,pos={180,445}	,size={70,53}	,title="Accept\r& Quit"				,help={"Keeps any changes made to the 2D data and closes the window."}
	
	TitleBox ContrastLabel		,pos={10,522}	,size={80,23}	,title="Image Contrast:"	,frame=0	,fStyle=1
	TitleBox ContrLowLabel		,pos={10,542}	,size={24,23}	,title="Min:"				,frame=0
	TitleBox ContrHighLabel		,pos={10,562}	,size={24,23}	,title="Max:"				,frame=0
	
	PopupMenu ContrastColor		,pos={120,520}	,size={140,20}	,title=""			,value= "*COLORTABLEPOPNONAMES*"
	Slider ContrastLow			,pos={48,547}	,size={210,10}	,side=0	,vert=0	,ticks=0
	Slider ContrastHigh			,pos={48,567}	,size={210,10}	,side=0	,vert=0	,ticks=0
	
	if (dim == 3)																				// add controls for 3D data
		//GroupBox SeparationLine3,pos={15,585}	,size={238,1}	,frame=0
		GroupBox LayersGroup	,pos={10,585}	,size={250,50}	,title="3D Layers"			,frame=0	,fstyle=1
		Button bDeleteLayer		,pos={120,605}	,size={50,23}	,title="Delete"				,help={"Deletes the selected layer from the stack. This can be undone."}
		Button bSumLayers		,pos={180,605}	,size={70,23}	,title="Merge All"			,help={"Merges layers of 3D data into one 2D image for processing. This can be undone."}
		SetVariable SetLayer	,pos={10,607}	,size={100,23}	,title="Layer:"				,help={"Selects the active layer of 3D data."}
		SetVariable SetLayer	,bodywidth = 55	,limits={0,DimSize(inwave,2)-1,1}	,value =_NUM:0	,proc=ChangeParameterControl
	endif
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pPath,";","*") fsize=10
	endif
	
	ModifyControlList ControlNameList(pPath, ";", "b*")  proc=Profile_MainButtonProc			// define the procedure for all buttons
	Button bNormUser			,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)			// activate measurement info normalization if the conditions are right
	Button bSaveSpec			,fColor=(65535,0,0)
	CheckBox IntColCheckBox		,value = 1 ,mode = 1	,proc=ModeCheckBoxControl				// radio button checked => horizontal mode
	CheckBox IntRowCheckBox		,value = 0 ,mode = 1	,proc=ModeCheckBoxControl				// radio button unchecked
	CheckBox FullWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MinWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox Normto1			,value = 0				,proc=ModeCheckBoxControl
	CheckBox SnapToGrid			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MaintainInt		,value = 0
	CheckBox cAlignSmooth		,value = 0
	CheckBox cAlignOutput		,value = 0
	CheckBox cAlignSubShift		,value = 1
	SetVariable Position		,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable Width			,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable SmoothFactor	,bodywidth = 55			,limits={0,32767,1}		,value=_NUM:(ceil(DimSize(inwave,0)*kAlignSmooth/100))
	PopupMenu ContrastColor		,bodyWidth = 140		,mode=1+WhichListItem("Terrain", CTabList())	,proc=Profile_ContrastColorProc
	
	Duplicate/Free/R=[][][0] inwave, tempLayer
	MatrixOp/free tempScale = replace(tempLayer,inf,0)											// make sure to remove infs, which mask the real maximum
	Variable wMin = WaveMin(tempScale), wMax = WaveMax(tempScale)
	Slider ContrastLow			,value=wMin				,limits={wMin,wMax,0}	,proc=Profile_ContrastSliderProc
	Slider ContrastHigh			,value=wMax				,limits={wMin,wMax,0}	,proc=Profile_ContrastSliderProc
	//+++++++++++++++++++++++++++++++++++++ set dependency +++++++++++++++++++++++++++++++++++++++
	Profile_AppendSectionTrace(pPath)															// add the ROI lines & set variable limits for the first time
	Profile_SetMinFullLineValues(pPath,1,0,0)													// set initial values for row mode (position & width)
	Profile_SetMinFullLineValues(pPath,1,1,0)													// set initial values for col mode

	Profile_UpdateLines(pPath)																	// first update to generate initial profile
	Profile_SetFullMinCheckbox(pPath)
	DoWindow/F $gName
	return 0
End

//################################################################################################

static Function/DF getWorkFolder(pPath)															// grab current working folder from panel info
	String pPath
	String folder = ""
	if (strlen(pPath))
		folder = GetUserData(pPath,"","workFolder")
	endif
	if (strlen(folder))
		return root:Packages:$(kWorkingDir):$(folder)
	else
		return root:Packages:$(kWorkingDir)														// backwards compatibility with older sessions
	endif
End

//################################################################################################

static Function/S getPanel(gName)
	String gName
	String subPanel = StringFromList(0,RemoveFromList("AskQuitPanel",ChildWindowList(gName)))	// grab attached panel from list
	if(strlen(subPanel))
		return gName+"#"+subPanel
	else																						// compatibility with older versions
		if (WinType("ProfileGeneratorWindow"))
			return "ProfileGeneratorWindow"
		else
			return ""
		endif
	endif
End

//################################################################################################

static Function/S getGraph(pPath)
	String pPath
	return StringFromList(0,pPath,"#")
End

//################################################################################################

static Function Profile_QuitProgram(inPath, Undo)
	String inPath
	Variable Undo
	
	String gName = getGraph(inPath)																// can be called from AskQuitPanel as well => construct panel path manually
	String pPath = getPanel(gName)
	DFREF stg = getWorkFolder(pPath)
	DFREF pkroot = getWorkFolder("")
	DFREF pkg = root:Packages:
	
	SetWindow $pPath hook(CreateProfile)=$""													// release hook to not trigger any events
	SetWindow $gName hook(CreateProfile)=$""
	KillWindow $getGraph(pPath)																	// first, close image window
	NVar/Z ProfileUpdateConnector = stg:ProfileUpdateConnector
	if (!NVAR_Exists(ProfileUpdateConnector))
		return -1																				// obviously the folder does not exist anymore
	endif
	SetFormula ProfileUpdateConnector,""														// reset formula
	KillDataFolder/Z root:WinGlobals
	
	if (Undo)																					// for some strange hook double call reason, the undo has to take place AFTER window reset
		SVAR wLoc = stg:InwaveLocation															// reference the modified wave
		Wave inwave = $wLoc
		Duplicate/O stg:StartCopy, inwave														// undo the changes before quitting
	endif
	
	if (CountObjectsDFR(stg, 4) == 0)															// backwards compatibility - makes sure to not kill the root folder with other open instances
		KillDataFolder/Z stg
	endif
	Variable Objects = CountObjectsDFR(pkroot, 1)+CountObjectsDFR(pkroot, 2)+CountObjectsDFR(pkroot, 3)+CountObjectsDFR(pkroot, 4)
	if (Objects == 0 && DataFolderRefStatus(pkroot) != 0)
		KillDataFolder pkroot
	endif
	Objects = CountObjectsDFR(pkg, 1)+CountObjectsDFR(pkg, 2)+CountObjectsDFR(pkg, 3)+CountObjectsDFR(pkg, 4)
	if (Objects == 0)																			// clean up Packages folder if empty
		KillDataFolder pkg
	endif
	return 0
End

//################################################################################################

static Function Profile_AppendSectionTrace(pPath)												// appends the profile parallel to the selection lines
	String pPath
	String gName = getGraph(pPath)
	
	DFREF stg = getWorkFolder(pPath)
	Wave Lines	= stg:CurrentLines
	Wave SecL	= stg:SectionLine
	
	Profile_SetPosWidthControls(pPath)															// initialize center and width controls
	
	String getSplit
	getSplit = GetUserData(pPath, "", "vertAxSplit")											// axes may have been dragged => get new split location
	Variable vertAx = strlen(getSplit) ? str2num(getSplit) : 0.3
	getSplit = GetUserData(pPath, "", "horiAxSplit")
	Variable horiAx = strlen(getSplit) ? str2num(getSplit) : 0.5
	
	RemoveFromGraph/W=$gName/Z CompareLine, SectionLine, MouseProfile							// remove all ROI traces
	If (Lines[0])	// vertical
		AppendToGraph/W=$gName/VERT/B=bottom2 SecL
		ModifyGraph/W=$gName axisEnab(left)={0,1},	axisEnab(bottom) ={vertAx+0.01,1}			// modify the main axes
		ModifyGraph/W=$gName freePos(bottom2)=0,	axisEnab(bottom2)={0,vertAx-0.01}			// align the new bottom axis
		SetAxis/W=$gName/A/E=1 bottom2															// scale from zero
	else		// horizontal
		AppendToGraph/W=$gName/L=left2 SecL
		ModifyGraph/W=$gName axisEnab(bottom)={0,1}, axisEnab(left) ={horiAx+0.01,1}
		ModifyGraph/W=$gName freePos(left2)=0,		 axisEnab(left2)={0,horiAx-0.01}
		SetAxis/W=$gName/A/E=1 left2
	endif
	ModifyGraph/W=$gName highTrip=100000
	ModifyGraph/W=$gName mirror=1, minor=1, standoff=0
	ModifyGraph/W=$gName ZisZ=1, notation=1, rgb=(65280,0,0)
	return 0
End

//################################################################################################

static Function Profile_SetPosWidthControls(pPath)
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	String unitStr = WaveUnits(inwave,1-Mode)
	Variable delta = abs(DimDelta(inwave,1-Mode))
	Variable init  = delta*round((DimSize(inwave,1-Mode)-1) / 25)								// calculate ca. 5% of width
	init = init <= delta ? 2*delta : init														// make sure that's not too small

	SetVariable Position ,win=$pPath ,limits={-inf,inf,init}		,value=Lines[1+Mode*2]	,format="%g "+unitStr	// set new control values	
	ControlInfo/W=$pPath SnapToGrid;	init = V_Value ? init+mod(init/delta,2)*delta : init	// make sure the change is even when snapped to the grid (=> symmetric width change)
	SetVariable Width	 ,win=$pPath ,limits={delta,inf,init}	,value=Lines[2+Mode*2]	,format="%g "+unitStr
	return 0
End

//################################################################################################

Function Profile_ProcessWindowEvent(s)															// main window hook
	STRUCT WMWinHookStruct &s

	Variable HookTakeover = 0
	if (WinType(s.WinName) == 7 && s.EventCode != 11)											// process only key presses for the panel
		return HookTakeover
	endif

	String gName = getGraph(s.winName)															// can be called from AskQuitPanel as well => construct panel path manually
	String pPath = getPanel(gName)
	if (!strlen(pPath))																			// no panel found
		return 0
	endif

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	if (s.EventCode > 2 && s.EventCode < 6 && CmpStr(gName,s.winName) == 0)						// for mouse events
		Variable axDivider, dragAxis = 0														// for a dynamic divider line
		sscanf StringByKey("axisEnab(x)", AxisInfo(gName, SelectString(Mode,"left","bottom")), "="), "{%f,", axDivider
		axDivider -= 0.01																		// center split position
		GetWindow $gName, psizeDC
		Variable plotRange	= Mode ? v_right - v_left : v_bottom - v_top
		Variable mouseClose = Mode ? abs((s.mouseloc.h-v_left)-axDivider*plotRange) < 5 : abs((s.mouseloc.v-v_top)-(1-axDivider)*plotRange) < 5
		Variable rMousePos	= Mode ? (s.mouseloc.h-v_left)/plotRange : 1-(s.mouseloc.v-v_top)/plotRange				// relative mouse position on the plot
	endif
	
	Switch (s.EventCode)
		case 17:	// window close button
			HookTakeover = 1
			if (!EqualWaves(stg:startCopy, inwave,-1))											// compare if the wave was changed
				if (!Wintype("#AskQuitPanel"))													// this will create a little panel in case of changes
					Variable pixConv = (CmpStr(IgorInfo(2), "Windows") == 0 && ScreenResolution == 96)? 72/ScreenResolution : 1
					Variable xCenter = (s.WinRect.right - s.WinRect.left)/2*72/ScreenResolution, yCenter = (s.WinRect.bottom - s.WinRect.top)/2*72/ScreenResolution
					NewPanel/N=AskQuitPanel/HOST=$gName/W=(xCenter-112*pixConv,yCenter-35*pixConv,xCenter+112*pixConv,yCenter+35*pixConv)
					DrawText 40,28,"Accept Changes to Data?"
					Button bAskOK		,pos={12,33}	,size={60,23}		,title="Yes"	,proc=Profile_AskButton
					Button bAskUndo		,pos={82,33}	,size={60,23}		,title="Undo"	,proc=Profile_AskButton
					Button bAskCancel	,pos={152,33}	,size={60,23}		,title="Cancel"	,proc=Profile_AskButton
					SetActiveSubwindow $gName													// restore focus to main window
				endif
			else
				 Profile_QuitProgram(pPath,0)													// else just quit (no undo)
			endif
		break
		case 11:	// key press
			HookTakeover = 1
			//++++++++++++++++++++++++++++++++ execute button keys ++++++++++++++++++++++++++++++
			String Key = ""
			Switch(s.keycode)
				case 112:	// p key
					Key = "bCompare"
					break
				case 115:	// s key
					Key = "bSaveSpec"
					break
				case 117:	// u key
					Key = "bUndoAll"
					break
				case 127:	// del key
					Key = "bRemoveSelection"
					break
			EndSwitch
			if (StrLen(Key) > 0)
				Profile_ExecuteButtonCommand(pPath, Key)
				break
			endif
			//++++++++++++++++++++++++++++++ execute checkbox keys ++++++++++++++++++++++++++++++
			Switch(s.keycode)
				case 11:	// page up
					Key = "FullWidth"
					break
				case 12:	// page down
					Key = "MinWidth"
					break
				case 103:	// g key
					Key = "toggleSnapToGrid"
					break
				case 104:	// h key
					Key = "IntColCheckBox"
					CheckBox IntColCheckBox ,value = 1 ,win=$pPath
					DrawAction/W=$gName getgroup=axisSplitLine, delete
					break
				case 109:	// m key
					Key = "toggleMousetoSpec"
					break
				case 110:	// n key
					Key = "toggleNormto1"
					break
				case 118:	// v key
					Key = "IntRowCheckBox"
					CheckBox IntRowCheckBox ,value = 1 ,win=$pPath
					DrawAction/W=$gName getgroup=axisSplitLine, delete
					break
			EndSwitch
			if (StrLen(Key) > 0)
				Profile_ExecuteCheckboxAction(pPath, Key,1)
				break
			endif
			//+++++++++++++++++++++++++++++++ execute arrow keys ++++++++++++++++++++++++++++++++
			ControlInfo/W=$pPath SnapToGrid
			Variable Shift = 0, Expand = 0, snapLines = V_Value
			Switch (Mode)																		// define a shift and expand direction (positive or negative) depending on mode and pressed key
				case 0:
					Shift	= s.keycode == 30 ? 1 : ( s.keycode == 31 ? -1 : Shift )
					Expand	= s.keycode == 29 ? 1 : ( s.keycode == 28 ? -1 : Expand )
					break
				case 1:
					Shift	= s.keycode == 29 ? 1 : ( s.keycode == 28 ? -1 : Shift )
					Expand	= s.keycode == 30 ? 1 : ( s.keycode == 31 ? -1 : Expand )
					break
			EndSwitch
			Variable Accelerate = 1																// add acceleration modifier depending on the pressed modifier key
			Accelerate = (GetKeyState(0) & 2^2) != 0 ? 4 : Accelerate
			Accelerate = (GetKeyState(0) & 2^0) != 0 ? 0.5 : Accelerate
			
			Variable Delta = abs(DimDelta(inwave,1 - Mode))
			Variable sizeMagn = ceil(DimSize(inwave,1 - Mode)/100)								// move further for bigger waves
			Lines[1 + Mode*2] += Shift * Delta * Accelerate * sizeMagn							// do the shift
			Lines[2 + Mode*2] += Expand* (1+snapLines) * Delta * Accelerate * sizeMagn			// do the expand (even expansion for snapped mode)
			Lines[2 + Mode*2] = Lines[2 + Mode*2] < Delta ? Delta : Lines[2 + Mode*2]			// check for the minimal width (will normally be done by the variable control)
			
			Profile_CheckBorders(pPath)															// check if the lines go over the borders of the wave
			Profile_SetFullMinCheckbox(pPath)													// toggle check-boxes
		break
		case 3: 	// mousedown
			if (mouseClose)																		// initialize axis drag state
				SetWindow $pPath userdata(axdragOn)="1"
				HookTakeover = 1
			endif
		break
		case 5: 	// mouseup
			SetWindow $pPath userdata(axdragOn)="0"												// release axis drag
		break
		case 4:		// mouse movement
			//++++++++++++++++++++++++++++++++ axis quick drag ++++++++++++++++++++++++++++++++++
			if (!WinType(pPath))																// catches slow window buildup in Igor 6
				break
			endif
			dragAxis = str2num(GetUserData(pPath, "", "axdragOn"))
			dragAxis = numtype(dragAxis) != 0 ? 0 : dragAxis
			if (dragAxis)
                GetMouse/W=$gName
                if (!(V_flag & 1))																// mouse button released? => maybe we missed a mouse-up event?
                    SetWindow $pPath userdata(axdragOn)="0"
                    dragAxis = 0
                endif
            endif
			DrawAction/W=$gName getgroup=axisSplitLine, delete									// draws and sets divider line for dynamic axis drag
			if (dragAxis)
				axDivider = limit(rMousePos, 0.15, 0.85)										// set new divider
				ModifyGraph/W=$gName axisEnab($SelectString(Mode,"left","bottom"))	 = {axDivider + 0.01, 1}
				ModifyGraph/W=$gName axisEnab($SelectString(Mode,"left2","bottom2")) = {0, axDivider - 0.01}
				SetWindow $pPath userdata($SelectString(Mode,"horiAxSplit","vertAxSplit"))=num2str(axDivider)		// save current position into panel
				HookTakeover = 1
			endif
			if (dragAxis || mouseClose)															// draw a line for visualization
				SetDrawEnv/W=$gName linefgc=(36000,15000,59000) ,dash=2 ,linethick= 1.50
				SetDrawEnv/W=$gName gstart, gname=axisSplitLine
				if (Mode)
					DrawLine/W=$gName axDivider, -0.1, axDivider, 1.1
				else
					DrawLine/W=$gName -0.1, 1-axDivider, 1.1, 1-axDivider
				endif
				SetDrawEnv/W=$gName gstop, gname=axisSplitLine
				s.cursorCode = Mode ? 5 : 6														// change mouse cursor
				s.doSetCursor = 1
			endif
			//++++++++++++++++++++++++++++++ quick mouse profile ++++++++++++++++++++++++++++++++
			ControlInfo/W=$pPath MousetoSpec
			if (V_Value)
				ControlInfo/W=$pPath SetLayer
				Variable layer = abs(V_flag) == 5 ? V_Value : 0									// get the current layer
				
				Variable Mx = round ((AxisValFromPixel("", "Bottom", s.mouseLoc.h)-DimOffset(inwave, 0)) / DimDelta(inwave, 0))
				Variable My = round ((AxisValFromPixel("", "Left", s.mouseLoc.v)-DimOffset(inwave, 1)) / DimDelta(inwave, 1))
				if (Mx >= DimSize(inwave, 0) || My >= DimSize(inwave, 1) || Mx < 0  || My < 0)
					break
				endif
				
				RemoveFromGraph/W=$gName/Z MouseProfile
				if (Mode)																		// append one row / column under the current mouse position
					AppendToGraph/W=$gName/VERT/B=bottom2 inwave[Mx][][layer]/TN=MouseProfile
				else
					AppendToGraph/W=$gName/L=left2 inwave[][My][layer]/TN=MouseProfile
				endif
				ControlInfo/W=$pPath Normto1													// also normalize the mouse profile
				if (V_Value)
					if (Mode)
						Duplicate/Free/R=[Mx][][layer] inwave, tempProfile
					else
						Duplicate/Free/R=[][My][layer] inwave, tempProfile
					endif
					MatrixOp/free tempScale = replace(tempProfile,inf,0)						// make sure to remove infs
					Variable wmin = wavemin(tempScale), wmax = wavemax(tempScale)
					ModifyGraph offset(MouseProfile)={,-wmin/(wmax-wmin)} ,muloffset(MouseProfile)={,1/(wmax-wmin)}
				endif
				ModifyGraph/W=$gName rgb(MouseProfile)=(0,39321,0)
			else
				RemoveFromGraph/W=$gName/Z MouseProfile
			endif
		break
	EndSwitch
	
	return HookTakeover
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function Profile_AskButton(s) : ButtonControl													// execute button commands of the quit sub-panel
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		Strswitch (s.ctrlName)
			case "bAskUndo":
				Profile_QuitProgram(s.win, 1)
				break
			case "bAskOK":
				Profile_QuitProgram(s.win, 0)
				break
			case "bAskCancel":																	// just remove the panel
				KillWindow #AskQuitPanel
				break
		EndSwitch
	endif
	return 0
End
	
//################################################################################################

static Function Profile_UpdateLines(pPath)														// generates marquee lines
	String pPath
	DFREF stg = getWorkFolder(pPath)
	Wave LinesX	= stg:LinesX
	Wave LinesY	= stg:LinesY
	Wave Lines	= stg:CurrentLines
	
	Variable plus, minus
	if (Lines[0] == 0)																			// depending on mode switch line direction
		plus	= Lines[1] + Lines[2]/2
		minus	= Lines[1] - Lines[2]/2
		LinesX	= {-INF,INF,NaN,-INF,INF}
		LinesY	= {plus,plus,NaN,minus,minus}
	else
		plus	= Lines[3] + Lines[4]/2
		minus	= Lines[3] - Lines[4]/2
		LinesX	= {plus,plus,NaN,minus,minus}
		LinesY	= {-INF,INF,NaN,-INF,INF}
	endif

	ModifyGraph/W=$getGraph(pPath) offset(LinesY)={0,0}											// reset offset, trigger S_TraceOffset
	return 0
End

//################################################################################################

Function LinesIntegrateUpdate(Info)																// will be triggered when S_TraceOffset changes (connected via formula)
	String Info
	if( StrSearch(Info,"TNAME:LinesY;",0) <=0 )													// contain the Info line data?
		return 0
	endif
	String gName = StringByKey("GRAPH",Info)
	String pPath = getPanel(gName)
	
	DFREF stg = getWorkFolder(pPath)
	if (DataFolderRefStatus(stg) == 0)															// something is wrong if the data is not found => exit
		return -1
	endif
	
	SVAR wLoc	= stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave SecPr	= stg:SectionLine
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	if (Mode == 0)																				// depending on mode (in row [0]) read the line data (X or Y direction)
		Lines[1] += NumberByKey("YOFFSET",Info)
	else
		Lines[3] += NumberByKey("XOFFSET",Info)
	endif

	Profile_CheckBorders(pPath)																	// check if the lines go over the borders of the wave before calculating something
	//++++++++++++++++++++++++++++++++++++ prepare integration ++++++++++++++++++++++++++++++++++
	ControlInfo/W=$pPath SetLayer
	Variable layer	= abs(V_flag) == 5 ? V_Value : 0
	Variable center	= Lines[1 + Mode * 2]
	Variable width	= Lines[2 + Mode * 2]
	Variable delta	= DimDelta(inwave, 1-Mode)
	Variable offset	= DimOffset(inwave, 1-Mode)
	Variable size	= DimSize(inwave, 1-Mode)

	Redimension/D/N=(DimSize(inwave, Mode)) SecPr												// set the right dimension and scale
	Variable start	= (center -	sign(delta) * width/2 - offset)/delta + 0.5						// recalculate to rows / columns x2pnt()-like; the sign(delta) reverses the direction for negative delta values
	Variable finish	= (center +	sign(delta) * width/2 - offset)/delta + 0.5						// shift the values by 0.5 to get the full rows /columns
	//start	= round(start/delta*1000)*delta/1000												// round the values depending on the delta size (to get rid of the floating point errors)
	//finish= round(finish/delta*1000)*delta/1000
	start	= start < 0 ? 0 : start																// validate values
	finish	= finish > size ? size : finish
	//Print/D "Start initial:",start,"Finish initial:",finish									// debug

	String NoteText = ""
	NoteText += note(inwave) + "\r" + NameOfWave(inwave)+" integrated from "+num2str(ChangePrecision(center-width/2,4))
	NoteText += " to "+num2str(ChangePrecision(center+width/2,4))+" "+WaveUnits(inwave,1-Mode)+ " (width = " +num2str(width)+" "+WaveUnits(inwave,1-Mode)+")"
	Note/K SecPr, NoteText																		// write comments
	
	Variable i, overhead = 0, rest = 0
	overhead = ceil(start) - start																// fraction at the beginning
	start	 = ceil(start)																		// beginning of the full columns / rows
	rest	 = finish - trunc(finish)															// rest from the end of the range
	finish	 = floor(finish)																	// maximal value
	
	if (start > finish)																			// if range is smaller than one column
		overhead += rest - 1
		rest = 0
	endif
	//Print/D "Overhead:",overhead,"Start:",start,"Finish:",finish,"Rest:",rest					// debug
	//++++++++++++++++++++++++++++++++++++ now integrate ++++++++++++++++++++++++++++++++++++++++
	SecPr = 0
	if (overhead > 0)
		SecPr += (Mode == 0 ? inwave[p][start-1][layer] : inwave[start-1][p][layer]) * overhead	// part of first line
	endif
	for (i = start; i < finish; i += 1)															// full lines
		SecPr += (Mode == 0 ? inwave[p][i][layer]		: inwave[i][p][layer])
	endfor
	if (rest > 0)
		SecPr += (Mode == 0 ? inwave[p][finish][layer]	: inwave[finish][p][layer]) * rest		// rest of the last line
	endif

	width /= abs(delta)																			// normalize for scaled waves
	if (width > 0)
		SecPr /= width																			// normalize to the integrated width
	endif
	//+++++++++++++++++++++++++++++++++++++++ normalize profile ++++++++++++++++++++++++++++++++++	
	ControlInfo/W=$pPath Normto1
	if (V_Value)																				// normalizes the spectrum to the maximum height
		start = waveMin(SecPr)
		finish = waveMax(SecPr)
		delta = finish-start
		delta = delta == 0 ? 1 : delta
		MatrixOP/O SecPr=(SecPr-start)/delta
	endif
	SetScale/P x, DimOffset(inwave,Mode), DimDelta(inwave,Mode), WaveUnits(inwave,Mode), SecPr	// set the scale last, after MatrixOP was executed
	return 0
End

//################################################################################################

static Function Profile_CheckBorders(pPath)														// check if the lines are within the boundaries, if not restrict
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable xdel = DimDelta(inwave,0), xoff = DimOffset(inwave,0)
	Variable ydel = DimDelta(inwave,1), yoff = DimOffset(inwave,1)
	Variable rows = DimSize(inwave,0) , cols = DimSize(inwave,1)
	Variable colPos	= Lines[1], colWidth= Lines[2]
	Variable rowPos	= Lines[3], rowWidth= Lines[4]
	//++++++++++++++++++++++++++ check for alignment with data grid ++++++++++++++++++++++++++++++
	Variable halfShift = 0
	ControlInfo/W=$pPath SnapToGrid
	if (V_Value)
		rowWidth	= xdel * round(rowWidth/xdel)
		colWidth	= ydel * round(colWidth/ydel)
		halfShift	= xdel/2*!mod(rowWidth/xdel,2)												// for even widths the center needs to be shifted over by half a row/col
		rowPos		= xoff + xdel*round((rowPos - halfShift - xoff)/xdel) + halfShift
		halfShift	= ydel/2*!mod(colWidth/ydel,2)
		colPos		= yoff + ydel*round((colPos - halfShift - yoff)/ydel) + halfShift
	endif
	//++++++++++++++++++++++++++++ read and check y position data ++++++++++++++++++++++++++++++++			
	Variable Boundary1 = yoff + (colWidth*sign(ydel) - ydel)/2
	Variable Boundary2 = yoff - (colWidth*sign(ydel) - ydel)/2 + ydel*(cols-1)
	
	colPos = (colPos - Boundary1)*sign(ydel) < 0 ? Boundary1 : colPos							// too low?
	colPos = (colPos - Boundary2)*sign(ydel) > 0 ? Boundary2 : colPos							// too high?
	Lines[1] = colPos																			// return correct value
	colWidth = colWidth > abs(ydel*cols) ? abs(ydel*cols) : colWidth							// too big?
	Lines[2] = colWidth
	//++++++++++++++++++++++++++++ read and check x position data ++++++++++++++++++++++++++++++++
	Boundary1 = xoff + (rowWidth*sign(xdel) - xdel)/2
	Boundary2 = xoff - (rowWidth*sign(xdel) - xdel)/2 + xdel*(rows-1)
	
	rowPos = (rowPos - Boundary1)*sign(xdel) < 0 ? Boundary1 : rowPos							// left check
	rowPos = (rowPos - Boundary2)*sign(xdel) > 0 ? Boundary2 : rowPos							// right check
	Lines[3] = rowPos
	rowWidth = rowWidth > abs(xdel*rows) ? abs(xdel*rows) : rowWidth							// too big?
	Lines[4] = rowWidth
	
	Profile_UpdateLines(pPath)																	// set the lines to the restricted values
	return 0
End

//#################################support functions for control elements#########################

Function ModeCheckBoxControl(s) : CheckBoxControl
	STRUCT WMCheckboxAction &s
	if(s.eventCode == 2)
		Profile_ExecuteCheckboxAction(s.win, s.ctrlName,s.checked)
	endif
	return 0
End
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function Profile_ExecuteCheckboxAction(pPath, type, checked)
	String pPath, type
	Variable checked
	
	DFREF stg  = getWorkFolder(pPath)
	Wave Lines = stg:CurrentLines
	StrSwitch (type)
		case "IntColCheckBox":
			Lines[0] = 0
			Profile_AppendSectionTrace(pPath)
			CheckBox IntRowCheckBox ,win=$pPath ,value = 0
		break
		case "IntRowCheckBox":
			Lines[0] = 1
			Profile_AppendSectionTrace(pPath)
			CheckBox IntColCheckBox ,win=$pPath ,value = 0
		break
		case "FullWidth":
			Profile_SetMinFullLineValues(pPath,1,Lines[0],checked)
		break
		case "MinWidth":
			Profile_SetMinFullLineValues(pPath,0,Lines[0],checked)
		break
		case "toggleNormto1":
			ControlInfo/W=$pPath Normto1
			CheckBox Normto1		,win=$pPath ,value = 1-V_Value
		break
		case "toggleMousetoSpec":
			ControlInfo/W=$pPath MousetoSpec
			CheckBox MousetoSpec	,win=$pPath ,value = 1-V_Value
		break
		case "toggleSnapToGrid":
			ControlInfo/W=$pPath SnapToGrid
			CheckBox SnapToGrid		,win=$pPath ,value = 1-V_Value
		case "SnapToGrid":
			Profile_SetPosWidthControls(pPath)
		break
	EndSwitch
	
	Profile_CheckBorders(pPath)																	// check if the lines go over the borders of the wave
	Profile_SetFullMinCheckbox(pPath)															// toggle check-boxes
	return 0
End

//################################################################################################

static Function Profile_SetMinFullLineValues(pPath, type, mode, activate)	
	String pPath
	Variable type, mode, activate																// type 0: min width; type 1: full width

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable size	= DimSize(inwave,1-mode)
	Variable offset	= DimOffset(inwave,1-mode)
	Variable delta	= DimDelta(inwave,1-mode)
	
	Variable init = abs(delta)*round((size-1) / 25)												// calculate ca. 5% of width
	init = init <= abs(delta) ? 2*abs(delta) : init												// make sure that's not too small
	if (type)
		Lines[1 + mode*2]  = offset																// center position
		Lines[1 + mode*2] += size < 100 ? delta*floor((size-1)/2) : delta*(size-1)/2			// align to grid for small waves
		Lines[1 + mode*2]  = abs(Lines[1 + mode*2]) < 10^-10*init ? 0 : Lines[1 + mode*2]		// properly center around zero
		Lines[2 + mode*2]  = activate == 1 ? abs(delta)*size : init								// full with or ca. 20 %
	else
		Lines[2 + mode*2]  = activate == 1 ? abs(delta) : init									// one line or 20 %
	endif
	return 0
End

//################################################################################################

Function ChangeParameterControl(s) : SetVariableControl											// execute variable changes and sets the check-boxes
	STRUCT WMSetVariableAction &s
	if (s.eventCode == 1 || s.eventCode == 2)
		if (CmpStr(s.ctrlName,"SetLayer") == 0)
			ModifyImage/W=$getGraph(s.win) ''#0 ,plane = s.dval									// replace the currently displayed layer
			Profile_UpdateImageContrast(s.win, 2)												// updates the contrast sliders for the new plane
		endif
		Profile_CheckBorders(s.win)																// check if the lines go over the borders of the wave
		Profile_SetFullMinCheckbox(s.win)														// set check-boxes correctly
	endif
	return 0
End
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function Profile_SetFullMinCheckbox(pPath)												// look, if a checkbox could be applied
	String pPath
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode  = Lines[0]
	Variable Delta = abs(DimDelta(inwave,1-Mode))
	CheckBox FullWidth	,win=$pPath	,value = Lines[2+Mode*2] >= Delta * DimSize(inwave,1-Mode)	// check for full width
	CheckBox MinWidth	,win=$pPath	,value = Lines[2+Mode*2] <= Delta							// check for minimal width
	return 0
End

//################################################################################################

Function Profile_ContrastSliderProc(s) : SliderControl
	STRUCT WMSliderAction &s
	if (s.eventCode & 1)
		ControlInfo/W=$s.win ContrastLow;	Variable minVal = V_Value
		ControlInfo/W=$s.win ContrastHigh;	Variable maxVal = V_Value
		ControlInfo/W=$s.win ContrastColor;	String color = S_Value
		ModifyImage/W=$getGraph(s.win) ''#0 ctab={minVal,maxVal,$color,0}
	endif
	return 0
End

//################################################################################################

Function Profile_ContrastColorProc(s) : PopupMenuControl
	STRUCT WMPopupAction &s
	if (s.eventCode == 2)
		ControlInfo/W=$s.win ContrastLow;	Variable minVal = V_Value
		ControlInfo/W=$s.win ContrastHigh;	Variable maxVal = V_Value
		ModifyImage/W=$getGraph(s.win) ''#0 ctab={minVal,maxVal,$s.popStr,0}
	endif
	return 0
End

//################################################################################################

Function Profile_MainButtonProc(s) : ButtonControl
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		Profile_ExecuteButtonCommand(s.win, s.ctrlName)
	endif
	return 0
End
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function Profile_ExecuteButtonCommand(pPath, which)
	String pPath, which

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable i, Mode = Lines[0]
	
	String logStr = ""
	String from	= num2str(ChangePrecision(Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2,4))		// current range
	String to	= num2str(ChangePrecision(Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2,4))
	
	ControlInfo/W=$pPath SetLayer
	Variable layer = abs(V_flag) == 5 ? V_Value : 0
	Variable wDim  = WaveDims(inwave)
	
	Variable intensitySum = 0																	// for later re-normalization
	Variable shiftDir = 1																		// shift direction starts positive
	Variable updateContrast = 0																	// updates the contrast sliders (1 = don't change range, 2 = reset range)
	
	Strswitch (which)
		case "bSaveQuit":
			Profile_QuitProgram(pPath, 0)
			return 0
		break
		case "bSaveSpec":
			logStr = ProfileAction_SaveProfile(pPath)
		break
		case "bCompare":
			ProfileAction_Compare(pPath)
		break
		case "bUndoAll":
			Profile_UndoLastUpdate(pPath, "all")
			logStr = "All changes undone."
			updateContrast = 2
		break
		case "bUndoLast":
			Profile_UndoLastUpdate(pPath, "last")
			logStr = "Last change undone."
			updateContrast = 2
		break
		case "bAlign":
			logStr = ProfileAction_Align(pPath)
			updateContrast = 1
		break
		case "bDeleteLayer":
			Profile_UndoLastUpdate(pPath, "save")
			
			DeletePoints/M=2 layer,1, inwave
			if (DimSize(inwave,2) < 2)															// deleted last extra layer
				ModifyImage/W=$getGraph(pPath) ''#0 ,plane = 0									// select the first layer and disable 3D controls
				SetVariable SetLayer ,win=$pPath	,disable = 2 ,value =_NUM:0
				Button bSumLayers	 ,win=$pPath	,disable = 2
				Button bDeleteLayer	 ,win=$pPath	,disable = 2
			else
				layer = layer > DimSize(inwave,2)-1 ? DimSize(inwave,2)-1 : layer
				ModifyImage/W=$getGraph(pPath) ''#0 ,plane = layer								// select the first layer and disable 3D controls
				SetVariable SetLayer ,win=$pPath	,value =_NUM:layer	,limits={0,DimSize(inwave,2)-1,1}
			endif
			logStr = "Deleted layer "+num2str(layer)+"."
		break
		case "bSumLayers":																		// merges 3D layers into an 2D image
			Profile_UndoLastUpdate(pPath, "save")
			
			ModifyImage/W=$getGraph(pPath) ''#0 ,plane = 0										// select the first layer and disable 3D controls
			SetVariable SetLayer ,win=$pPath	,disable = 2	,value =_NUM:0
			Button bSumLayers	 ,win=$pPath	,disable = 2
			ControlInfo/W=$pPath bDeleteLayer													// backwards compatibility - added delete button in update
			if (V_flag == 1)
				Button bDeleteLayer	,win=$pPath	,disable = 2
			endif
			
			MatrixOP/Free tMat = sumBeams(inwave)/waveLayers(inwave)							// merge layers
			CopyScales inwave, tMat;	Note tMat, note(inwave)									// copy scales and notes over
			Duplicate/O tMat, inwave
			
			logStr = "Merged all layers into 2D image."
			updateContrast = 2
		break
		case "bNormalize":																		// normalize to profile shape
			Profile_UndoLastUpdate(pPath, "save")
			MatrixOP/Free tempSec=replace(SecPr,0,inf)											// the profile should not get zero => have a big number instead to suppress this row
			if( !((GetkeyState(0) & 2^0) != 0) && kNormSmooth > 0)								// ctrl / command = no smoothing
				Smooth round(kNormSmooth), tempSec												// smooth the profile a bit
			endif
			for (i = 0; i < DimSize(inwave,1-Mode); i += 1)
				if (Mode)
					inwave[i][][layer] /= tempSec[q]
				else
					inwave[][i][layer] /= tempSec[p]	
				endif
			endfor
			intensitySum = mean(SecPr)
			
			logStr = "Normalized to "+SelectString((GetkeyState(0) & 2^0),"","(exact) ") + SelectString(Mode,"horizontal","vertical")
			logStr += " profile (Range: "+from+"-"+to+")" + SelectString(wDim>2,""," of layer "+num2str(layer)) + "."
			updateContrast = 2
		break
		case "bNormMax":																		// normalize to maximum
			Profile_UndoLastUpdate(pPath, "save")
			Make/FREE/D/N=(DimSize(inwave,1-Mode)) tempwave										// get a temporary wave
			for (i = 0; i < DimSize(inwave,Mode); i += 1)
				tempwave = Mode > 0 ? inwave[p][i][layer] : inwave[i][p][layer]					// extract the row / column
				Variable maxV = Wavemax(tempwave)
				maxV = maxV == 0 ? 1 : maxV
				if (Mode)
					inwave[][i][layer] /= maxV
				else
					inwave[i][][layer] /= maxV
				endif
				intensitySum += maxV															// add the values for re-normalization later
			endfor
			intensitySum /= i+1
			logStr = "Normalized to each "+SelectString(Mode,"vertical","horizontal") + " line maximum" + SelectString(wDim>2,""," of layer "+num2str(layer)) + "."
			updateContrast = 2
		break
		case "bNormUser":																		// normalize to measurement value
			Profile_UndoLastUpdate(pPath, "save")
			intensitySum = Profile_ExecuteUserInfoButton(inwave)
			Button bNormUser ,win=$pPath ,disable= 2											// disable the button to avoid double clicking (more than once is meaningless anyway)			
			logStr = "Normalized by "+kUserInfoBtnName+"."
			updateContrast = 2
		break
		case "bSubtractPr":		
			Profile_UndoLastUpdate(pPath, "save")
			ProfileAction_SubtractProfile(pPath)
			logStr = SelectString(Mode,"Horizontal","Vertical") + " profile (Range: "+from+"-"+to+") subtracted" + SelectString(wDim>2,""," from layer "+num2str(layer)) + "."
			updateContrast = 1
		break
		case "bRemoveSelection":
			Profile_UndoLastUpdate(pPath, "save")
			logStr = ProfileAction_RemoveSelection(pPath)
			updateContrast = 1
		break
		case "bShiftYAxisMinus":
			shiftDir = -1
		case "bShiftYAxisPlus":
			Profile_UndoLastUpdate(pPath, "save")
			Profile_ShiftButtonStatus(pPath, shiftDir)											// enable and disable shift buttons
			ProfileAction_ShiftImage(pPath, shiftDir)
			logStr = "Data shifted to "+SelectString(shiftDir+1,"negative","positive")+" y-axis values."
			updateContrast = 1
		break
	Endswitch
	
	ControlInfo/W=$pPath MaintainInt
	if (V_Value && intensitySum != 0 && numtype(intensitySum) == 0)								// restore the average height of the spectrum (from the average height of the Profile)
		inwave[][][layer] *= intensitySum
		logStr += " Average intensity is maintained."
	endif
	
	ControlInfo/W=$pPath cPrintEdits															// print logs if desired
	if (V_flag == 2 && V_Value == 1 && strlen(logStr))
		logStr[0] = NameOfWave(inwave) + " (" + time() + "): "									
		Print logStr
	endif
	
	Profile_UpdateImageContrast(pPath, updateContrast)
	Profile_UpdateLines(pPath)
	return 0
End

//################################################################################################

static Function Profile_UpdateImageContrast(pPath, updateMode)
	String pPath
	Variable updateMode
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave inwave	= $wLoc

	ControlInfo/W=$pPath ContrastColor
	if (updateMode > 0 && abs(V_flag) == 3)
		String color = S_Value
		Variable minVal, maxVal
		
		ControlInfo/W=$pPath SetLayer
		Variable layer = abs(V_flag) == 5 ? V_Value : 0
		
		Duplicate/Free/R=[][][layer] inwave, tempLayer
		MatrixOp/free tempScale = replace(tempLayer,inf,0)
		Variable wMin = WaveMin(tempScale), wMax = WaveMax(tempScale)
		
		if (updateMode == 1)																	// mode 1 reads current setting from the contrast sliders
			ControlInfo/W=$pPath ContrastLow
			minVal = limit(V_Value,wMin,wMax)
			ControlInfo/W=$pPath ContrastHigh
			maxVal = limit(V_Value,wMin,wMax)
			minVal = minVal >= maxVal ? wMin : minVal
		else
			minVal = wMin
			maxVal = wMax
		endif
		
		Slider ContrastLow	,win=$pPath	,value=minVal	,limits={wMin,wMax,0}
		Slider ContrastHigh	,win=$pPath	,value=maxVal	,limits={wMin,wMax,0}
		ModifyImage/W=$getGraph(pPath) ''#0 ctab={minVal,maxVal,$color,0}
		
		// if (minVal == wMin && maxVal != wMax)												// this interferes with 3D data where the contrast is layer-based and is not needed
			// ModifyImage/W=$getGraph(pPath) ''#0 ctab={*,maxVal,$color,0}
		// elseif (minVal != wMin && maxVal == wMax)
			// ModifyImage/W=$getGraph(pPath) ''#0 ctab={minVal,*,$color,0}
		// elseif (minVal == wMin && maxVal == wMax)
			// ModifyImage/W=$getGraph(pPath) ''#0 ctab={*,*,$color,0}
		// else
			// ModifyImage/W=$getGraph(pPath) ''#0 ctab={minVal,maxVal,$color,0}
		// endif
	endif
	return 0
End

//################################################################################################

static Function/S ProfileAction_SaveProfile(pPath)
	String pPath

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave SecPr	= stg:SectionLine
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]

	String outName, name = NameOfWave(inwave), unit = WaveUnits(inwave,0)
	Variable skipEnd = (GetkeyState(0) & 2^2) != 0												// enables a keyboard shortcut to just cut the name from the ending
	
	String from	= num2str(ChangePrecision(Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2,4))		// current range
	String to	= num2str(ChangePrecision(Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2,4))
	
	ControlInfo/W=$pPath FullWidth
	if (V_Value || skipEnd)
		outName	= name+"_pr"+SelectString(Mode , "H", "V")
	else
		if (IgorVersion() < 7.00 && (strlen(name) + strlen(from) + strlen(to) + 5) > 31 )		// make sure to not go over the 31 character limit
			from = from[0,5]
			to	 = to[0,5]
			Variable chars = 25 - strlen(from) - strlen(to)
			outName = name[0,chars] + SelectString(Mode, "_H(" , "_V(") + from+"-"+to + ")"
		else
			outName = name + SelectString(Mode, "_H(" , "_V(") + from+" to "+to + ")"
		endif
	endif
	Duplicate/O SecPr, $outName
	
	String logStr = "Saved "+SelectString(Mode,"horizontal","vertical")+" profile (integrated from "
	logStr += from+" "+unit+" to "+to+" "+unit+") => output to "+outName+"."
	return logStr
End
//++++++++++++++++++++++++++++++++++++++++ number rounding +++++++++++++++++++++++++++++++++++++++
static Function ChangePrecision(value, numSigDigits)											// round values to select-able number of digits
	variable value, numSigDigits
	string str
	sprintf str, "%.*g\r", numSigDigits, value
	return str2num(str)
End

//################################################################################################

static Function ProfileAction_Compare(pPath)
	String pPath
	String gName = getGraph(pPath)
	DFREF stg = getWorkFolder(pPath)
	Wave SecPr = stg:SectionLine
	Wave Lines = stg:CurrentLines
	RemoveFromGraph/W=$gName/Z CompareLine
	Duplicate/O SecPr, stg:CompareLine															// duplicate current profile for comparison
	if (Lines[0] == 1)																			// append together with the main profile
		AppendToGraph/W=$gName/VERT/B=bottom2 stg:CompareLine
	else
		AppendToGraph/W=$gName/L=left2 stg:CompareLine
	endif
	ModifyGraph/W=$gName rgb(CompareLine)=(0,0,52224)
	return 0
End

//################################################################################################

static Function/S ProfileAction_RemoveSelection(pPath)
	String pPath
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable Mode	= Lines[0]
	Variable from	= Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2								// calculate the position of the Lines
	Variable to		= Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2
	Variable offset = DimOffset(inwave,1-Mode), delta = DimDelta(inwave,1-Mode)
	Variable nOff	= offset-(from-to)															// recalculate the offset (is used if the removed section is at the left / bottom side)
	//Print "start:",from,"stop:", to, "offset:", nOff											// debug (range in wave scales)
	from = round((from + delta/2 - offset)/delta)												// which points do the lines include
	to	 = round((to + delta/2 - offset)/delta)
	//Print "start:",from,"stop:", to, "count:", abs(to - from)									// debug (range in wave points)
	
	DeletePoints/M=(1-Mode) from, abs(to-from), Inwave											// delete the selected points
	if (from == 0)																				// shifts the scale only if the wave is cut at the old offset
		if (Mode)
			SetScale/P x nOff, delta, WaveUnits(inwave,1-Mode),  inwave
		else
			SetScale/P y nOff, delta, WaveUnits(inwave,1-Mode),  inwave
		endif
	endif
	
	String logStr = SelectString(Mode,"Horizontal","Vertical")+" ROI section deleted ("
	logStr += num2str(abs(to-from))+" lines in point range "+num2str(from)+"-"+num2str(to)+")."
	return logStr
End

//################################################################################################

static Function ProfileAction_ShiftImage(pPath, shiftDir)
	String pPath
	Variable shiftDir
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave inwave = $wLoc
	Wave/Z lastcpy = stg:LastCopy																// read wave data from the copy if available
	if (!WaveExists(lastcpy))
		Wave lastcpy = $wLoc
	endif
	
	Variable modkey	= (GetKeyState(0) & 2^0) != 0												// when the 'ctrl/command' key is pressed, the excess amount will be cut from the front of the spectrum	
	Variable rows = DimSize(lastcpy,0),  cols = DimSize(lastcpy,1)
	Variable xdel = DimDelta(lastcpy,0), xoff = DimOffset(lastcpy,0)
	Variable ydel = DimDelta(lastcpy,1), yoff = DimOffset(lastcpy,1)
	
	Variable addrow = ceil((cols-1)*abs(ydel/xdel))												// how many rows need to be added for the shift
	Variable nOff	= xoff + shiftDir*yoff														// the new x-offset
	nOff += sign(ydel)<0 ? shiftDir*(cols-1)*ydel : 0											// for negative y-deltas add the full width (basically start from the bottom)
	if (modkey && (addrow < rows-10))															// delete incomplete parts
		Redimension/D/N=(rows-addrow,-1,-1) inwave
		nOff += shiftDir ==  1*sign(xdel) ? addrow*xdel : 0
	else
		Redimension/D/N=(rows+addrow,-1,-1) inwave
		nOff -= shiftDir == -1*sign(xdel) ? sign(xdel)*(cols-1)*abs(ydel) : 0					// for negative shifts the start offset needs to be shifted as well (by the y-width of the spectrum)
	endif
	SetScale/P x nOff, xdel, WaveUnits(inwave,0), inwave

	Variable i
	if (WaveDims(inwave) == 3)																	// only shift the current layer of a 3D image
		for (i = 0; i < DimSize(inwave,2); i += 1)
			Duplicate/Free/R=[][][i] lastcpy, tLayer
			Redimension/N=(-1,-1,0) tLayer
			inwave[][][i] = interp2D(tLayer, nOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel)	// calculate every shifted point
		endfor
		MatrixOP/Free tMat=replaceNaNs(inwave,0)												// fill NaNs with zero
		inwave = tMat
	else
		inwave[][] = interp2D(lastcpy, nOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel)
		MatrixOP/O inwave=replaceNaNs(inwave,0)
	endif
	//+++++++++++++++++++++++ if there are zero rows, remove them ++++++++++++++++++++++++
	for (i = 0; i < DimSize(inwave,0); i += 1)
		MatrixOp/Free partSum = Sum(Row(inwave,i))
		if (partSum[0] != 0)
			break
		endif
	endfor
	if (i > 0)
		DeletePoints 0,i, inwave
		SetScale/P x DimOffset(inwave,0)+i*DimDelta(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0),  inwave
	endif
	
	Variable excessRows = 0, lastData = -1
	for (i = DimSize(inwave,0)-1; i > 0; i -= 1)
		MatrixOp/Free partSum = Sum(Row(inwave,i))
		if (partSum[0] == 0)
			excessRows += 1
		else
			if (lastData == -1)
				lastData = i
			endif
		endif
	endfor
	if (excessRows > 0 && lastData != -1)
		DeletePoints lastData+1, excessRows, inwave
	endif
	
	return 0
End

//################################################################################################

static Function Profile_ShiftButtonStatus(pPath, mode)											// disables the button after one press to prevent multiple shifts
	String pPath
	Variable mode	// -1 = -Y; 0 = undo; 1 = +Y;
	Variable currMode = 0																		// first, find the current disable state of the two buttons
	ControlInfo/W=$pPath bShiftYAxisPlus;	currMode = (V_disable == 2) ?  1 : currMode
	ControlInfo/W=$pPath bShiftYAxisMinus;	currMode = (V_disable == 2) ? -1 : currMode
	Variable Yplus  = currMode == 0 && mode ==  1 ? 2 : 0										// deactivate the relevant button
	Variable Yminus = currMode == 0 && mode == -1 ? 2 : 0
	Button bShiftYAxisPlus  ,win=$pPath ,disable=Yplus
	Button bShiftYAxisMinus ,win=$pPath ,disable=Yminus
	Button bUndoLast		,win=$pPath ,userdata(lastShift)=num2str(mode)
	return 0
End

//################################################################################################

static Function ProfileAction_SubtractProfile(pPath)
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable Mode = Lines[0]
	Variable xOff = DimOffset(inwave,0), xDel = DimDelta(inwave,0)
	Variable yOff = DimOffset(inwave,1), yDel = DimDelta(inwave,1)
	Variable size = DimSize(inwave,1-Mode), i
	
	ControlInfo/W=$pPath SetLayer
	Variable layer = abs(V_flag) == 5 ? V_Value : 0
	
	Variable LeftOrRight = 0
	LeftOrRight = GetKeyState(0) == 1 ?  1 : LeftOrRight										// ctrl/command for positive shift
	LeftOrRight = GetKeyState(0) == 2 ? -1 : LeftOrRight										// alt/option for negative shift
	if (Mode == 0 && LeftOrRight != 0)															// subtracts along the Y axis (increasing/decreasing with Y values)
		Variable xadd  = ceil(size*abs(yDel/xDel) )												// Add points at both sides to stay within bounds
		Variable xsize = DimSize(inwave,0)
		Variable nOff  = xOff-xDel*xadd - LeftOrRight*(Lines[1]-yOff)							// new offset: add extended range and offset from current Y-center (Lines[1])
		Make/D/Free/N=(xsize+2*xadd) tempcut
		SetScale/P x, nOff, xDel ,tempcut														// readjust scale for additional points
		tempcut[0,xadd-1]					= SecPr[0]											// fill with extended profile (endpoints will be repeated)
		tempcut[xadd,xsize+xadd-1]			= SecPr[p-xadd]
		tempcut[xsize+xadd,xsize+2*xadd-1]	= SecPr[xsize-1]
	endif

	for (i = 0; i < size; i += 1)
		if (Mode == 1)
			inwave[i][][layer] -= SecPr[q]
		else
			if(WaveExists(Tempcut))
				inwave[][i][layer] -= tempcut(x-LeftOrRight*yDel*i)								// subtract a moving profile
			else
				inwave[][i][layer] -= SecPr[p]													// subtract stationary profile
			endif
		endif
	endfor
	return 0
End

//################################################################################################

static Function/S ProfileAction_Align(pPath)
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	DFREF sav = GetDataFolderDFR()
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	Wave/Z backup = stg:TempCopy																// create an extra save of the current state
	if (!WaveExists(backup))
		Duplicate inwave, stg:TempCopy
		Wave backup = stg:TempCopy
	endif
	
	String name = NameOfWave(inwave)
	Variable size = DimSize(inwave,1-Mode)
	Variable line = DimSize(inwave,Mode)

	ControlInfo/W=$pPath SetLayer;			Variable layer = abs(V_flag) == 5 ? V_Value : 0
	ControlInfo/W=$pPath cAlignSmooth;		Variable doSmth = V_Value == 1
	ControlInfo/W=$pPath cAlignOutput;		Variable output = V_Value == 1
	ControlInfo/W=$pPath cAlignEdges;		Variable edgeAlign = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pPath cAlignSubShift; 	Variable subShift = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pPath SmoothFactor;		Variable smthFact = abs(V_Flag) == 5 ? V_Value : ceil(line*kAlignSmooth/100)

	Variable rev = (GetkeyState(0) & 2^2) != 0 ? 1 : 0											// shift = reverse sorting 0 = no, 1 = yes	
	String logStr = SelectString(subShift,"Aligned ","Subs-shift aligned ")+SelectString(edgeAlign,"","edges ")+"to the "
	
	Variable chars = IgorVersion() < 7.00 ? 25: inf
	if (output)																					// writes align shift into a file
		Make/O/D/N=(size) $(name[0,chars]+"_shft")
		Wave out = $(name[0,chars]+"_shft")
		SetScale/P x DimOffset(inwave,1-Mode), DimDelta(inwave,1-Mode), WaveUnits(inwave,1-Mode), out	
	endif

	Variable from = 0, to = line-1, count														// use the cursors to select a sub-region for alignment
	if (strlen(CsrInfo(A)) > 0)
		from = Mode ? qcsr(A,"") : pcsr(A,"")
		from = numtype(from) == 0 ? from : 0
	endif
	if (strlen(CsrInfo(B)) > 0)
		to = Mode ? qcsr(B,"") : pcsr(B,"")
		to = numtype(to) == 0 ? to : line-1
	endif
	if (from > to)
		count = from
		from  = to
		to	  = count
	endif
	
	Make/FREE/N=(line) ShiftWorkA = 0, ShiftWorkB = 0, tempWave = 0
	if (Mode)
		ShiftWorkA[from,to] = backup[0 + rev*(size-1)][p][layer]
		logStr += SelectString(rev,"left","right") + " side"
	else
		ShiftWorkA[from,to] = backup[p][0 + rev*(size-1)][layer]
		logStr += SelectString(rev,"bottom","top")
	endif
	if(doSmth && smthFact > 0)
		Smooth smthFact, ShiftWorkA
		logStr += " with smoothing (factor " + num2str(smthFact) + ")"
	endif
	if (output)
		logStr += " => output to " + NameOfWave(out)
	endif
	if (edgeAlign)																				// align the derivative instead
		Differentiate ShiftWorkA/D=tempWave
		ShiftWorkA = tempWave
	endif
	
	if (subShift)
		DebuggerOptions
		Variable sav_debug = V_debugOnError
		DebuggerOptions debugOnError=0 															// switch this off in case the fit fails
	endif
	
	SetDataFolder stg
	inwave[][][layer] = NaN
	for (count = 0; count < size; count += 1)													// find the correlation and shift for each row/column
		if (Mode)
			ShiftWorkB = backup[count][p][layer]
		else
			ShiftWorkB = backup[p][count][layer]
		endif
		if (subShift)
			Duplicate/Free ShiftWorkB, currLine
		endif
		if(doSmth && smthFact > 0)
			Smooth smthFact, ShiftWorkB
		endif
		if (edgeAlign)																			// align the derivative instead
			Differentiate ShiftWorkB/D=tempWave
			ShiftWorkB = tempWave
		endif
		if (from > 0)
			ShiftWorkB[0,from] = 0
		endif
		if (to < line-1)
			ShiftWorkB[to,line-1] = 0
		endif
		
		Variable align = SpectrumCorrelationShift(ShiftWorkA,ShiftWorkB, subShift)
		ShiftWorkB = NaN
		if (Mode)
			if (subShift)		
				ShiftWorkB[-floor(align)*(align < 0),line-1 -ceil(align)*(align >= 0)] = currLine(x+align)
				inwave[count][][layer] = ShiftWorkB[q]											// write interpolated result
			else
				inwave[count][-align*(align < 0),line-1 -align*(align >= 0)][layer] = backup[count][q+align][layer]
			endif
		else
			if (subShift)
				ShiftWorkB[-floor(align)*(align < 0),line-1 -ceil(align)*(align >= 0)] = currLine(x+align)
				inwave[][count][layer] = ShiftWorkB[p]
			else
				inwave[-align*(align < 0),line-1 -align*(align >= 0)][count][layer] = backup[p+align][count][layer]
			endif
		endif
		
		if(output)
			out[count] = DimDelta(inwave, Mode) * align
		endif
	endfor
	SetDataFolder sav
	
	if (subShift)
		DebuggerOptions debugOnError=sav_debug
	endif
	
	return logStr+"."
End

//+++++++++++++++++++++++++++++++++++++ correlation function +++++++++++++++++++++++++++++++++++++

static Function SpectrumCorrelationShift(in1,in2,sub)											// Helper function: Finds maximum correlation of two slices
	Wave in1,in2																				// data to compare
	Variable sub
	
	Duplicate/Free in2, work
	Correlate/AUTO in1, work
	WaveStats/Q work
	Variable error = 1, result = 0																// result = 0 means perfect match
	Variable optLoc = V_maxRowLoc - x2pnt(work, 0)
	if (V_sum != 0)
		if (V_maxRowLoc >= 3 && sub)
			try
				CurveFit/M=0/Q gauss, work[V_maxRowLoc-3,V_maxRowLoc+3]; AbortOnRTE
				Wave W_coef
				if (numtype(W_coef[2]) == 0 && W_coef[2] > optLoc-3 && W_Coef[2] < optLoc+3)
					result = W_coef[2]
					error = 0
				endif
			catch
				if (V_AbortCode == -4)
					Variable cfError = GetRTError(1)											// 1 to clear the error
					// Print "Error during sub-step alignment:", GetErrMessage(cfError,3)		// debug
				endif
			endtry
		endif
		if (error)
			result = optLoc
		endif
	endif
	return result
End

//################################################################################################

static Function Profile_UndoLastUpdate(pPath, which)
	String pPath, which

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave startcopy	= stg:StartCopy
	Wave/Z lastcopy = stg:LastCopy
	Wave/Z tempcopy	= stg:TempCopy
	Wave inwave		= $wLoc

	StrSwitch(which)
		case "save":
			Duplicate/O inwave, stg:LastCopy
		break
		case "last":
			if (WaveExists(tempcopy))
				Duplicate/O tempcopy, inwave, lastcopy											// updating lastcopy will prevent loading an older version upon second press
			elseif (WaveExists(lastcopy))
				Duplicate/O lastcopy, inwave
			else
				Duplicate/O startcopy, inwave
			endif
			
			String last = GetUserData(pPath,"bUndoLast","lastShift")							// checks if shift button should be restored
			if (strlen(last) > 0)
				Profile_ShiftButtonStatus(pPath, -str2num(last))
			endif
		break
		case "all":
			Duplicate/O startcopy, inwave
			KillWaves/Z lastcopy
			Profile_ShiftButtonStatus(pPath, 0)
		break
	EndSwitch
	KillWaves/Z tempcopy
	
	ControlInfo/W=$pPath SetLayer
	if (abs(V_flag) == 5)
		Variable upperLimit = DimSize(inwave,2)-1
		upperLimit = upperLimit < 0 ? 0 : upperLimit
		if (upperLimit > 0)
			SetVariable SetLayer	,win=$pPath ,disable = 0	,limits={0,upperLimit,1}
			Button bSumLayers		,win=$pPath	,disable = 0
			ControlInfo/W=$pPath bDeleteLayer													// backwards compatibility - added delete button in update
			if (V_flag == 1)
				Button bDeleteLayer	,win=$pPath	,disable = 0
			endif
		endif
	endif
	Button bUndoLast ,win=$pPath ,userdata(lastShift)=""										// delete shift-button info
	Button bNormUser ,win=$pPath ,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)			// re-enable the button if data is available
	return 0
End